import typing
from typing import AnyStr, Dict, List, Tuple, Union
from requests.structures import CaseInsensitiveDict
from pyramid.config import Configurator
from pyramid.registry import Registry
from pyramid.request import Request as PyramidRequest
from pyramid.response import Response as PyramidResponse
from webob.response import Response as WebobResponse
from webob.headers import ResponseHeaders, EnvironHeaders

if hasattr(typing, "TypedDict"):
    from typing import TypedDict  # pylint: disable=E0611,no-name-in-module
else:
    from typing_extensions import TypedDict  # noqa

Number = Union[int, float]
AnyValue = Union[AnyStr, Number, bool, None]
AnyKey = Union[AnyStr, int]
JSON = Dict[AnyKey, Union[AnyValue, Dict[AnyKey, 'JSON'], List['JSON']]]

AnyContainer = Union[Configurator, Registry, PyramidRequest]
SettingValue = Union[AnyStr, Number, bool, None]
SettingsType = Dict[AnyStr, SettingValue]
AnySettingsContainer = Union[AnyContainer, SettingsType]

CookiesType = Dict[AnyStr, AnyStr]
HeadersType = Dict[AnyStr, AnyStr]
CookiesTupleType = List[Tuple[AnyStr, AnyStr]]
HeadersTupleType = List[Tuple[AnyStr, AnyStr]]
CookiesBaseType = Union[CookiesType, CookiesTupleType]
HeadersBaseType = Union[HeadersType, HeadersTupleType]
OptionalHeaderCookiesType = Union[Tuple[None, None], Tuple[HeadersBaseType, CookiesBaseType]]
AnyHeadersContainer = Union[HeadersBaseType, ResponseHeaders, EnvironHeaders, CaseInsensitiveDict]
AnyCookiesContainer = Union[CookiesBaseType, PyramidRequest, AnyHeadersContainer]
AnyResponseType = Union[WebobResponse, PyramidResponse]



from twitcher.interface import OWSRegistryInterface
from twitcher.utils import sanitize

import logging
LOGGER = logging.getLogger("TWITCHER")


class OWSRegistry(OWSRegistryInterface):
    """
    OWS Service Registry is a service to register OWS services for the OWS proxy.
    """
    def __init__(self, servicestore):
        self.store = servicestore

    def register_service(self, name, url, *args, **kwargs):
        """
        Adds an OWS service with the given ``name`` and ``url`` to the service store.
        """
        data = dict(kwargs)
        data['name'] = sanitize(name)
        data['url'] = url
        try:
            service = self.store.save_service(**data)
        except Exception:
            LOGGER.exception('register service failed')
            return {}
        return service.json()

    def unregister_service(self, name):
        """
        Removes OWS service with the given ``name`` from the service store.
        """
        try:
            self.store.delete_service(name=name)
        except Exception:
            LOGGER.exception('unregister service failed')
            return False
        else:
            return True

    def get_service_by_name(self, name):
        """
        Gets service with given ``name`` from service store.
        """
        try:
            service = self.store.fetch_by_name(name=name)
        except Exception as exc:
            msg = 'Could not get service with name {}'.format(name)
            LOGGER.debug(msg, exc_info=exc)
            LOGGER.error(msg)
            return {}
        else:
            return service.json()

    def get_service_by_url(self, url):
        """
        Gets service with given ``url`` from service store.
        """
        try:
            service = self.store.fetch_by_url(url=url)
        except Exception:
            LOGGER.error('Could not get service with url {}'.format(url))
            return {}
        else:
            return service.json()

    def list_services(self):
        """
        Lists all registered OWS services.
        """
        try:
            services = [service.json() for service in self.store.list_services()]
        except Exception:
            LOGGER.error('List services failed.')
            return []
        else:
            return services

    def clear_services(self):
        """
        Removes all services from the service store.
        """
        try:
            self.store.clear_services()
        except Exception:
            LOGGER.error('Clear services failed.')
            return False
        else:
            return True


def includeme(config):
    from twitcher.adapter import get_adapter_factory

    def owsregistry(request):
        adapter = get_adapter_factory(request)
        return adapter.owsregistry_factory(request)

    # In case the adapter employs caching or other per-request/session dependent transaction details, we must ensure
    # to regenerate the owsregistry object each time since the service-store it provides (amongst other things),
    # only initializes the request once instead of per-request method calls.
    # For example, 'request.owsregistry.get_service_by_name' will call 'ServiceStore.fetch_by_name' with
    # the 'ServiceStore' initialized and stored with the first ever request (if reify=True). All following service
    # operations would employ the stored database session contained within this request.
    config.add_request_method(owsregistry, reify=False, property=True)



"""
Factories to create storage backends.
"""

from twitcher.adapter.base import AdapterInterface
from twitcher.models.service import ServiceConfig
from twitcher.owsproxy import send_request
from twitcher.owssecurity import OWSSecurity
from twitcher.owsregistry import OWSRegistry
from twitcher.store import ServiceStore
from twitcher.utils import get_settings
from pyramid.config import Configurator
from pyramid.request import Request
from pyramid.response import Response

TWITCHER_ADAPTER_DEFAULT = 'default'


class DefaultAdapter(AdapterInterface):
    @property
    def name(self):
        return TWITCHER_ADAPTER_DEFAULT

    def describe_adapter(self):
        from twitcher.__version__ import __version__
        return {"name": self.name, "version": str(__version__)}

    def configurator_factory(self, container):
        settings = get_settings(container)
        return Configurator(settings=settings)

    def owssecurity_factory(self):
        return OWSSecurity()

    def owsregistry_factory(self, request):
        return OWSRegistry(ServiceStore(request))

    def owsproxy_config(self, container):
        from twitcher.owsproxy import owsproxy_defaultconfig
        # update provided config or generate it otherwise
        if not isinstance(container, Configurator):
            container = self.configurator_factory(container)
        owsproxy_defaultconfig(container)

    def request_hook(self, request, service):
        return request

    def response_hook(self, response, service):
        return response

    def send_request(self, request: Request, service: ServiceConfig) -> Response:
        return send_request(request, service)



"""
The ``twitcherctl`` is a command line tool to control the twitcher service.
It is used to generate access tokens and to register OWS services.

``twitcherctl`` is part of the twitcher installation:

.. code-block:: console

   $ twitcherctl -h

`twitcherctl` Commands and Options
----------------------------------

``twitcherctl`` has the following command line options:

-h, --help

   Print usage message and exit

-s, --serverurl

   URL on which twitcher server is listening (default "http://localhost:8000/").

-u, --username

   Username to access twitcher server.

-p, --password

   Password to access twitcher server.

-k, --insecure

   Don't validate the server's certificate.

List of available commands:

add
    Add OAuth2 client application.
gentoken
    Generates an access token.
list
    Lists all registered OWS services used by OWS proxy.
clear
    Removes all OWS services from the registry.
register
   Adds OWS service to the registry to be used by the OWS proxy.
unregister
   Removes OWS service from the registry.

Add an OAuth2 client application
--------------------------------

Register an application using basic authentication with username and password
(name and redirect-uri are optional):

.. code-block:: console

    $ twitcherctl -k --username demo --password demo add --name demo_app --redirect-uri http://localhost/demo_app
    {'client_id': 'id', 'client_secret': 'secret'}

The result is an OAuth *client_id* and *client_secret*.

Generate an access token
------------------------

Get an OAuth access token using *client_id* and *client_secret* for given *scope*:

.. code-block:: console

   $ twitcherctl -k gentoken -i client_id -s client_secret --scope compute
   {'access_token': 'TOKEN', 'expires_in': 3600, 'scope': ['compute'], 'token_type': 'Bearer'}

Possible scopes are: compute, register.

You can also get a token from a Keycloak OAuth service using the client credentials workflow:

.. code-block:: console

   $ twitcherctl -k -s http://localhost:8080 gentoken -i client_id -s client_secret --scope compute --keycloak

Register an OWS Service for the OWS Proxy
-----------------------------------------

See the available options:

.. code-block:: console

   twitcherctl -k register -h

Register a local WPS service using an OAuth access token:

.. code-block:: console

   $ twitcherctl -k --username demo --password demo register http://localhost:5000/wps
   tiny_buzzard

You can use the ``--name`` option to provide a name (used by the OWS proxy).
Otherwise a nice name will be generated.

List registered services
------------------------

The ``list`` command shows the registered OWS services:

.. code-block:: console

   $ twitcherctl -k --username demo --password demo list
   [{'url': 'http://localhost:5000/wps', 'type': 'wps', 'name': 'tiny_buzzard', 'auth': 'token'}]

"""

import sys
import getpass
import argcomplete
import argparse

from twitcher.client import TwitcherService
from twitcher.namesgenerator import get_random_name

import logging
logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.WARN)
LOGGER = logging.getLogger("TWITCHER")


class TwitcherCtl(object):
    """
    Command line to interact with the OAuth and OpenAPI interface of the ``twitcher`` service.
    """

    def create_parser(self):
        parser = argparse.ArgumentParser(
            prog="twitcherctl",
            description='twitcherctl -- control twitcher service from the cmd line.',
        )
        parser.add_argument("--debug",
                            help="Enable debug mode.",
                            action="store_true")
        parser.add_argument('-s', '--serverurl',
                            metavar='URL',
                            default='http://localhost:8000',
                            help='URL on which twitcher server is listening (default "http://localhost:8000").')
        parser.add_argument("-u", "--username",
                            help="Username to use for authentication with server.")
        parser.add_argument("-p", "--password",
                            help="Password to use for authentication with server")
        parser.add_argument("-k", "--insecure",  # like curl
                            help="Don't validate the server's certificate.",
                            action="store_true")

        # commands
        subparsers = parser.add_subparsers(
            dest='cmd',
            title='command',
            description='List of available commands',
        )

        # add client app
        # --------------

        # add
        subparser = subparsers.add_parser('add', help="Add OAuth2 client application")
        subparser.add_argument("-n", "--name",
                               help="Client application name")
        subparser.add_argument("-r", "--redirect-uri",
                               help="Client application redirect URI")

        # token management
        # ----------------

        # gentoken
        subparser = subparsers.add_parser('gentoken', help="Generates an OAuth2 access token.")
        subparser.add_argument("-i", "--client-id", help="OAuth client-id")
        subparser.add_argument("-s", "--client-secret", help="OAuth client-secret")
        subparser.add_argument("-S", "--scope", nargs='+', default='compute', help="OAuth scope: compute or register")
        subparser.add_argument("-K", "--keycloak",
                               help="Use keycloak token endpoint.",
                               action="store_true")

        # service registry
        # ----------------

        # list
        subparser = subparsers.add_parser('list', help="Lists all registered OWS services used by OWS proxy.")

        # clear
        subparser = subparsers.add_parser('clear', help="Removes all OWS services from the registry.")

        # register
        subparser = subparsers.add_parser('register',
                                          help="Adds OWS service to the registry to be used by the OWS proxy.")
        subparser.add_argument('url', help="Service url.")
        subparser.add_argument('--name', help="Service name. If not set then a name will be generated.")
        subparser.add_argument('--type', default='wps',
                               help="Service type (wps, wms). Default: wps.")
        subparser.add_argument('--purl', default='',
                               help="Service optional public URL.")
        subparser.add_argument('--auth', default='token',
                               help="Authentication method (token, cert, public). Default: token.")
        subparser.add_argument('--verify', default='true',
                               help="Verify SSL service certificate (true, false). Default: true.")

        # unregister
        subparser = subparsers.add_parser('unregister', help="Removes OWS service from the registry.")
        subparser.add_argument('name', help="Service name.")

        return parser

    def run(self, args):
        if args.debug:
            LOGGER.setLevel(logging.DEBUG)

        if args.insecure:
            # See: https://urllib3.readthedocs.io/en/latest/advanced-usage.html#ssl-warnings
            import urllib3
            urllib3.disable_warnings()
            LOGGER.warning('disabled certificate verification!')

        username = password = None
        if args.cmd != 'gentoken':
            username = args.username
            if not username:
                username = input('Username: ')
            password = args.password
            if not password:
                password = getpass.getpass(prompt='Password: ')

        verify_ssl = args.insecure is False
        service = TwitcherService(
            url=args.serverurl,
            username=username,
            password=password,
            verify=verify_ssl)

        try:
            if args.cmd == 'list':
                return service.list_services()
            elif args.cmd == 'register':
                data = {'type': args.type,
                        'purl': args.purl,
                        'auth': args.auth,
                        'verify': args.verify}
                return service.register_service(
                    name=args.name or get_random_name(),
                    url=args.url,
                    data=data,
                )
            elif args.cmd == 'unregister':
                return service.unregister_service(name=args.name)
            elif args.cmd == 'clear':
                return service.clear_services()
            elif args.cmd == 'add':
                return service.add_client_app(
                    name=args.name,
                    redirect_uri=args.redirect_uri)
            elif args.cmd == 'gentoken':
                client_id = args.client_id
                if not client_id:
                    client_id = input('Client ID: ')
                client_secret = args.client_secret
                if not client_secret:
                    client_secret = getpass.getpass(prompt='Client Secret: ')
                return service.fetch_token(
                    client_id=client_id,
                    client_secret=client_secret,
                    scope=args.scope,
                    keycloak=args.keycloak)
        except Exception as e:
            LOGGER.error("Error: {}".format(e))
        return None


def main():
    LOGGER.setLevel(logging.INFO)

    ctl = TwitcherCtl()
    parser = ctl.create_parser()
    argcomplete.autocomplete(parser)

    args = parser.parse_args()
    return ctl.run(args)


if __name__ == '__main__':
    sys.exit(main())



from pyramid.config import Configurator
from pyramid.request import Request
from pyramid.response import Response

from twitcher.interface import OWSSecurityInterface, OWSRegistryInterface
from twitcher.models.service import ServiceConfig
from twitcher.typedefs import AnySettingsContainer, JSON
from twitcher.utils import get_settings


class AdapterInterface(object):
    """
    Common interface allowing functionality overriding using an adapter implementation.
    """
    def __init__(self, container: AnySettingsContainer) -> None:
        self.settings = get_settings(container)

    @property
    def name(self):
        return '{}.{}'.format(self.__module__, type(self).__name__)

    def describe_adapter(self) -> JSON:
        """
        Returns a JSON serializable dictionary describing the adapter implementation.
        """
        raise NotImplementedError

    def configurator_factory(self, container: AnySettingsContainer) -> Configurator:
        """
        Returns the 'configurator' implementation of the adapter.
        """
        raise NotImplementedError

    def owssecurity_factory(self) -> OWSSecurityInterface:
        """
        Returns the 'owssecurity' implementation of the adapter.
        """
        raise NotImplementedError

    def owsregistry_factory(self, request: Request) -> OWSRegistryInterface:
        """
        Returns the 'owsregistry' implementation of the adapter.
        """
        raise NotImplementedError

    def owsproxy_config(self, container: AnySettingsContainer) -> None:
        """
        Returns the 'owsproxy' implementation of the adapter.
        """
        raise NotImplementedError

    def request_hook(self, request: Request, service: ServiceConfig) -> Request:
        """
        Apply modifications onto the request before sending it.

        .. versionadded:: 0.7.0

        Request members employed after this hook is called include:
        - :meth:`Request.headers`
        - :meth:`Request.method`
        - :meth:`Request.body`

        This method can modified those members to adapt the request for specific service logic.
        """
        raise NotImplementedError

    def response_hook(self, response: Response, service: ServiceConfig) -> Response:
        """
        Apply modifications onto the response from sent request.

        .. versionadded:: 0.7.0

        The received response from the proxied service is normally returned directly.
        This method can modify the response to adapt it for specific service logic.
        """
        raise NotImplementedError

    def send_request(self, request: Request, service: ServiceConfig) -> Response:
        """
        Performs the provided request in order to obtain a proxied response.

        .. versionadded:: 0.8.0

        The operation should consider the service definition to resolve where the
        request redirection should be proxied to, and handle any relevant response
        errors, such as an unauthorized access or an unreachable service.
        """
        raise NotImplementedError



from twitcher.namesgenerator import get_random_name


def test_get_random_name():
    name = get_random_name()
    assert len(name) > 3
    assert '_' in name


def test_get_random_name_retry():
    name = get_random_name(retry=True)
    assert len(name) > 3
    assert int(name[-1]) >= 0



from .base import FunctionalTest


def compute_view(request):
    if request.verify_request(scopes=["compute"]) is True:
        result = {'status': 'success'}
    else:
        result = {'status': 'not allowed'}
    return result


class OAuth2AppTests(FunctionalTest):

    def setUp(self):
        super(OAuth2AppTests, self).setUp()
        self.init_database()

        self.config.include('twitcher.oauth2')
        self.config.add_route('compute', '/api/compute')
        self.config.add_view(compute_view, route_name='compute', renderer='json')
        self.app = self.get_test_app()

    def test_compute_with_param(self):
        access_token = self.create_token()
        resp = self.app.get('/api/compute?access_token={}'.format(access_token))
        assert resp.status_code == 200
        assert resp.content_type == 'application/json'
        resp.mustcontain('success')

    def test_compute_with_header(self):
        access_token = self.create_token()
        resp = self.app.get('/api/compute',
                            headers={'Authorization': 'Bearer {}'.format(access_token)})
        assert resp.status_code == 200
        assert resp.content_type == 'application/json'
        resp.mustcontain('success')



"""
The OWSRequest is based on pywps code:

* https://github.com/geopython/pywps/tree/pywps-3.2/pywps/Parser
* https://github.com/geopython/pywps/blob/master/pywps/app/WPSRequest.py
"""

import lxml.etree

from pyramid.httpexceptions import HTTPBadRequest
from twitcher.owsexceptions import (
    OWSNoApplicableCode,
    OWSInvalidParameterValue,
    OWSMissingParameterValue)
from twitcher.utils import lxml_strip_ns

allowed_service_types = ('wps', 'wms')
allowed_request_types = {'wps': ('getcapabilities', 'describeprocess', 'execute', 'getstatus', 'getresult'),
                         'wms': ('getcapabilities',
                                 'getmap',
                                 'getfeatureinfo',
                                 'getlegendgraphic',
                                 # ncwms extras,
                                 'getmetadata')}
public_request_types = {'wps': ('getcapabilities', 'describeprocess', 'getstatus', 'getresult'),
                        'wms': ('getcapabilities', )}
allowed_versions = {'wps': ('1.0.0', '2.0.0'), 'wms': ('1.1.1', '1.3.0',)}


class OWSRequest(object):
    """
    ``OWSRequest`` parses on OWS request and provides methods to access the parameters.
    """

    def __init__(self, request):
        self.parser = ows_parser_factory(request)
        self.parser.parse()

    @property
    def service(self):
        return self.parser.params['service']

    @property
    def request(self):
        # TODO: same name for service request and HTTP request is confusing.
        return self.parser.params['request']

    @property
    def version(self):
        return self.parser.params['version']

    def service_allowed(self):
        return self.service in allowed_service_types

    def public_access(self):
        return self.request in public_request_types[self.service]


def ows_parser_factory(request):
    if request.method == 'GET':
        return Get(request)
    elif request.method == 'POST':
        return Post(request)
    else:
        raise HTTPBadRequest()


class OWSParser(object):

    def __init__(self, request):
        self.request = request
        self.params = {}

    def parse(self):
        self._get_service()
        self._get_request_type()
        self._get_version()
        return self.params

    def _get_service(self):
        raise NotImplementedError

    def _get_request_type(self):
        raise NotImplementedError

    def _get_version(self):
        raise NotImplementedError


class Get(OWSParser):

    def _request_params(self):
        new_params = {}
        for param in self.request.params:
            # new_params[param.lower()] = self.request.params.getone(param)
            new_params[param.lower()] = self.request.params[param]
        return new_params

    def _get_param(self, param, allowed_values=None, optional=False):
        """Get parameter in GET request."""
        request_params = self._request_params()
        if param in request_params:
            value = request_params[param].lower()
            if allowed_values is not None:
                if value in allowed_values:
                    self.params[param] = value
                else:
                    raise OWSInvalidParameterValue("%s %s is not supported" % (param, value), value=param)
        elif optional:
            self.params[param] = None
        else:
            raise OWSMissingParameterValue('Parameter "%s" is missing' % param, value=param)
        return self.params[param]

    def _get_service(self):
        """Check mandatory service name parameter in GET request."""
        return self._get_param(param="service", allowed_values=allowed_service_types)

    def _get_request_type(self):
        """Find requested request type in GET request."""
        return self._get_param(param="request", allowed_values=allowed_request_types[self.params['service']])

    def _get_version(self):
        """Find requested version in GET request."""
        version = self._get_param(param="version", allowed_values=allowed_versions[self.params['service']],
                                  optional=True)
        if version is None and self._get_request_type() != "getcapabilities":
            raise OWSMissingParameterValue('Parameter "version" is missing', value="version")
        else:
            return version


class Post(OWSParser):

    def __init__(self, request):
        super(Post, self).__init__(request)

        try:
            xml = self.request.body
            self.document = lxml.etree.fromstring(xml)
            lxml_strip_ns(self.document)
        except Exception as e:
            raise OWSNoApplicableCode("{}".format(e))

    def _get_service(self):
        """Check mandatory service name parameter in POST request."""
        if "service" in self.document.attrib:
            value = self.document.attrib["service"].lower()
            if value in allowed_service_types:
                self.params["service"] = value
            else:
                raise OWSInvalidParameterValue("Service %s is not supported" % value, value="service")
        else:
            raise OWSMissingParameterValue('Parameter "service" is missing', value="service")
        return self.params["service"]

    def _get_request_type(self):
        """Find requested request type in POST request."""
        value = self.document.tag.lower()
        if value in allowed_request_types[self.params['service']]:
            self.params["request"] = value
        else:
            raise OWSInvalidParameterValue("Request type %s is not supported" % value, value="request")
        return self.params["request"]

    def _get_version(self):
        """Find requested version in POST request."""
        if "version" in self.document.attrib:
            value = self.document.attrib["version"].lower()
            if value in allowed_versions[self.params['service']]:
                self.params["version"] = value
            else:
                raise OWSInvalidParameterValue("Version %s is not supported" % value, value="version")
        elif self._get_request_type() == "getcapabilities":
            self.params["version"] = None
        else:
            raise OWSMissingParameterValue('Parameter "version" is missing', value="version")
        return self.params["version"]



# -*- coding: utf-8 -*-

# This information is located in its own file so that it can be loaded
# without importing the main package when its dependencies are not installed.
# See: https://packaging.python.org/guides/single-sourcing-package-version
__doc__ = """Security Proxy for OGC Services like WPS."""
__author__ = """Carsten Ehbrecht"""
__email__ = 'ehbrecht@dkrz.de'
__version__ = '0.11.0'



# coding: utf-8
"""
twitcher.oauth2
---------------

This module implements the `OAuth2 <https://oauthlib.readthedocs.io/en/latest/oauth2/oauth2.html>`_
model to generate access tokens.
These tokens are used to access the compute services (*scope=compute*).
The compute services are accessed via the OWS proxy.

The implementation is using the OAuth2
`client credentials <https://oauthlib.readthedocs.io/en/latest/oauth2/grants/credentials.html>`_
grant type

Currently three types of access tokens can be used:

*random_token*
    The access token is a UUID string.
    The tokens are stored in a local database and can be used for local validation only.

*signed_token*
    A JWT token signed with a X.509 certificate.
    The token can be validated without contacting a validation service.

*custom_token*
    A JWT token with a secret (UUID string) which can be shared for validation.
    The token can be validated without contacting a validation service.

*keycloak_token*
    A JWT token generated by a `Keycloak <https://www.keycloak.org/>`_ OAuth2 service.

See also the OAuth2
`token documenation <https://oauthlib.readthedocs.io/en/latest/oauth2/tokens/tokens.html>`_

The implementation is using the `pyramid-oauthlib <https://github.com/tilgovi/pyramid-oauthlib/>`_ library.
The code is also inspired by the following OAuth libraries:

* https://github.com/thomsonreuters/bottle-oauthlib
* https://github.com/lepture/flask-oauthlib

Further reading:

* https://oauthlib.readthedocs.io/en/latest/index.html
* https://pypi.org/project/PyJWT/
* https://requests-oauthlib.readthedocs.io/en/latest/index.html
* https://www.slideshare.net/alvarosanchezmariscal/stateless-authentication-with-oauth-2-and-jwt-javazone-2015
* https://docs.apigee.com/api-platform/security/oauth/oauth-20-client-credentials-grant-type
"""

import datetime
import uuid

from oauthlib.oauth2 import RequestValidator
from oauthlib.oauth2.rfc6749 import tokens
import jwt

from pyramid.settings import asbool

from twitcher import models
from twitcher.utils import get_settings

import logging
LOGGER = logging.getLogger('TWITCHER')


TOKEN_ENDPOINT = '/oauth/token'
CLIENT_APP_ENDPOINT = '/oauth/client'
DEFAULT_SCOPES = ['compute']


class Client():
    client_id = None


class BaseValidator(RequestValidator):
    default_grants = ["client_credentials"]

    def _get_client(self, request, client_id):
        query = request.dbsession.query(models.Client)
        client = query.filter(models.Client.client_id == client_id).first()
        return client

    def get_default_scopes(self, client_id, request, *args, **kwargs):
        LOGGER.debug('get_default_scopes: client_id={}'.format(client_id))
        client = self._get_client(request, request.client_id)
        if client:
            return client.default_scopes
        else:
            return []

    def authenticate_client(self, request, *args, **kwargs):
        LOGGER.debug('authenticate_client: {}'.format(request.client_id))
        client = self._get_client(request, request.client_id)
        if not client:
            return False
        request.client = Client()
        request.client.client_id = request.client_id
        request.user = request.client_id
        if request.client_secret == client.client_secret:
            return True
        return False

    def validate_grant_type(self, client_id, grant_type, client, request, *args, **kwargs):
        return grant_type in self.default_grants

    def validate_scopes(self, client_id, scopes, client, request, *args, **kwargs):
        LOGGER.debug('validate_scopes: client_id={}, scopes={}'.format(
            client_id, scopes))
        _client = self._get_client(request, client_id)
        if not _client:
            return False
        return all(scope in _client.default_scopes for scope in scopes)

    def save_bearer_token(self, token_response, request, *args, **kwargs):
        pass


class RandomTokenValidator(BaseValidator):
    def save_bearer_token(self, token_response, request, *args, **kwargs):
        """Persist the Bearer token."""
        token = models.Token(client_id=request.client_id, **token_response)
        request.dbsession.add(token)

    def generate_access_token(self, request):
        return tokens.random_token_generator(request)

    def validate_bearer_token(self, token, scopes, request):
        """Validate access token.

        :param token: A string of random characters
        :param scopes: A list of scopes
        :param request: The Request object passed by oauthlib

        The validation validates:

            1) if the token is available
            2) if the token has expired
            3) if the scopes are available
        """
        query = request.dbsession.query(models.Token)
        tok = query.filter(models.Token.access_token == token).first()
        if not tok:
            return False
        # validate expires
        if tok.expires is not None and \
                datetime.datetime.utcnow() > tok.expires:
            return False
        # validate scopes
        if scopes and not set(tok.scopes) & set(scopes):
            return False
        return True


class SignedTokenValidator(BaseValidator):
    def __init__(self, cert, key, issuer):
        self.cert = cert
        self.key = key
        self.issuer = issuer

    def generate_access_token(self, request):
        private_pem_key = open(self.key, "br").read()
        return tokens.signed_token_generator(private_pem_key, issuer=self.issuer)(request)

    def validate_bearer_token(self, token, scopes, request):
        public_pem = open(self.cert, "br").read()
        return tokens.common.verify_signed_token(public_pem, token)


class CustomTokenValidator(BaseValidator):
    def __init__(self, secret, issuer):
        self.secret = secret
        self.issuer = issuer

    def generate_access_token(self, request):
        token = jwt.encode({
            "ref": str(uuid.uuid4()),
            # "aud": request.client_id,
            "iss": self.issuer,
            "exp": datetime.datetime.utcnow() + datetime.timedelta(seconds=request.expires_in)
        }, self.secret, algorithm='HS256')
        return token

    def validate_bearer_token(self, token, scopes, request):
        try:
            jwt.decode(token, self.secret, verify=True, algorithms=['HS256'])
        except Exception:
            return False
        else:
            return True


class KeycloakTokenValidator(BaseValidator):
    def __init__(self, secret):
        self.public_key = '-----BEGIN PUBLIC KEY-----\n{}\n-----END PUBLIC KEY-----'.format(secret)

    def generate_access_token(self, request):
        raise NotImplementedError("This validator can only validate tokens.")

    def validate_bearer_token(self, token, scopes, request):
        try:
            jwt.decode(token, self.public_key, audience='account', verify=True, algorithms=['RS256'])
        except Exception as e:
            LOGGER.debug('token validation failed: {}'.format(e))
            return False
        else:
            return True


def generate_token_view(request):
    """Core functionality is available directly from the request.

    Responses from OAuthLib are wrapped in a response object of type
    :class:`pyramid.response.Response` so they can be returned directly
    from views.
    """
    # Extra credentials we need in the validator
    # credentials = {'user': request.user}
    LOGGER.debug('generate_token_view: client_id={}, client_secret={}'.format(
        request.client_id, request.client_secret))
    return request.create_token_response(credentials=None)


def register_client_app_view(request):
    """
    Register a new client application and returns ``client_id`` and ``client_secret``.

    Uses basic authentication.
    """
    client = models.Client(
        name=request.params.get('name'),
        client_id=uuid.uuid4().hex,
        client_secret=uuid.uuid4().hex,
        _redirect_uris=request.params.get('redirect_uri'),
        default_scope=DEFAULT_SCOPES[0])
    request.dbsession.add(client)
    return dict(
        name=client.name,
        client_id=client.client_id,
        client_secret=client.client_secret,
        redirect_uri=client.default_redirect_uri,
        scope=client.default_scopes)


def includeme(config):
    settings = get_settings(config)
    if asbool(settings.get('twitcher.oauth', True)):
        config.include('pyramid_oauthlib')
        # using basic auth for client app registration
        config.include('twitcher.basicauth')

        # Validator callback functions are passed Pyramid request objects so
        # you can access your request properties, database sessions, etc.
        # The request object is populated with accessors for the properties
        # referred to in the OAuthLib docs and used by its built in types.
        token_type = settings.get('twitcher.token.type', 'random_token')
        if token_type == 'random_token':
            validator = RandomTokenValidator()
        elif token_type == 'signed_token':
            validator = SignedTokenValidator(
                cert=settings.get('twitcher.token.certfile'),
                key=settings.get('twitcher.token.keyfile'),
                issuer=settings.get('twitcher.token.issuer'))
        elif token_type == 'custom_token':
            validator = CustomTokenValidator(
                secret=settings.get('twitcher.token.secret'),
                issuer=settings.get('twitcher.token.issuer'))
        elif token_type == 'keycloak_token':
            validator = KeycloakTokenValidator(
                secret=settings.get('keycloak.token.secret'))
        else:  # default
            validator = RandomTokenValidator()

        # Register grant types to validate token requests.
        config.add_grant_type('oauthlib.oauth2.ClientCredentialsGrant',
                              request_validator=validator)

        # Register the token types to use at token endpoints.
        config.add_token_type('oauthlib.oauth2.BearerToken',
                              request_validator=validator,
                              token_generator=validator.generate_access_token,
                              expires_in=int(settings.get('twitcher.token.expires_in', '3600')))

        config.add_route('access_token', TOKEN_ENDPOINT)
        config.add_view(generate_token_view, route_name='access_token')
        config.add_route('client', CLIENT_APP_ENDPOINT)
        config.add_view(register_client_app_view, route_name='client', renderer='json', permission='view')



from pyramid.config import Configurator
from pyramid.request import Request
from pyramid.registry import Registry
from datetime import datetime
from urllib import parse as urlparse
from lxml import etree
import json
import time
import pytz
import re
from typing import AnyStr, Optional

from twitcher.exceptions import ServiceNotFound
from twitcher.typedefs import AnySettingsContainer, SettingsType

import logging
LOGGER = logging.getLogger("TWITCHER")


def get_settings(container: AnySettingsContainer) -> Optional[SettingsType]:
    """
    Retrieves the application ``settings`` from various containers referencing to it.

    :raises TypeError: if the container type cannot be identified to retrieve settings.
    """
    if isinstance(container, (Configurator, Request)):
        container = container.registry
    if isinstance(container, Registry):
        container = container.settings
    if isinstance(container, dict):
        return container
    raise TypeError("Could not retrieve settings from container object of type [{}]".format(type(container)))


def get_twitcher_url(container: AnySettingsContainer) -> AnyStr:
    settings = get_settings(container)
    return settings.get('twitcher.url').rstrip('/').strip()


def sanitize(name, minlen=2, maxlen=25):
    """Lower-case name and replace all non-ascii chars by `_`."""
    if name is None or len(name.strip()) < minlen:
        raise ValueError("name must have at least {} chars.".format(minlen))
    return re.sub(r'\W|^(?=\d)', '_', name.strip().lower()[:maxlen])


def is_valid_url(url):
    try:
        parsed_url = urlparse.urlparse(url)
        return True if all([parsed_url.scheme, ]) else False
    except Exception:
        return False


def is_json_serializable(item):
    try:
        json.dumps(item)
        return True
    except (TypeError, OverflowError):
        return False


def parse_service_name(url: str, protected_path: str) -> Optional[str]:
    parsed_url = urlparse.urlparse(url)
    service_name = None
    if parsed_url.path.startswith(protected_path):
        parts_without_protected_path = parsed_url.path[len(protected_path)::].strip('/').split('/')
        # use ranges to avoid index error in case the path parts list is empty
        # the expected part must be exactly the first one after the protected path, then followed by the service name
        if any(part in parts_without_protected_path[:1] for part in ['proxy', 'verify']):
            parts_without_protected_path = parts_without_protected_path[1:]
        if len(parts_without_protected_path) > 0:
            service_name = parts_without_protected_path[0]
    if not service_name:
        raise ServiceNotFound
    return service_name


def now():
    return localize_datetime(datetime.utcnow())


def now_secs():
    """
    Return the current time in seconds since the Epoch.
    """
    return int(time.time())


def expires_at(hours=1):
    return now_secs() + hours * 3600


def localize_datetime(dt, tz_name='UTC'):
    """Provide a timezone-aware object for a given datetime and timezone name
    """
    tz_aware_dt = dt
    if dt.tzinfo is None:
        utc = pytz.timezone('UTC')
        aware = utc.localize(dt)
        timezone = pytz.timezone(tz_name)
        tz_aware_dt = aware.astimezone(timezone)
    else:
        LOGGER.warning('tzinfo already set')
    return tz_aware_dt


def baseurl(url):
    """
    return baseurl of given url
    """
    parsed_url = urlparse.urlparse(url)
    if not parsed_url.netloc or parsed_url.scheme not in ("http", "https"):
        raise ValueError('bad url')
    service_url = "%s://%s%s" % (parsed_url.scheme, parsed_url.netloc, parsed_url.path.strip())
    return service_url


def path_elements(path):
    elements = [el.strip() for el in path.split('/')]
    elements = [el for el in elements if len(el) > 0]
    return elements


def lxml_strip_ns(tree):
    for node in tree.iter():
        try:
            has_namespace = node.tag.startswith('{')
        except AttributeError:
            continue  # node.tag is not a string (node is a comment or similar)
        if has_namespace:
            node.tag = node.tag.split('}', 1)[1]


def replace_caps_url(xml, url, prev_url=None):
    ns = {
        'ows': 'http://www.opengis.net/ows/1.1',
        'xlink': 'http://www.w3.org/1999/xlink'}
    doc = etree.fromstring(xml)
    # wms 1.1.1 onlineResource
    if 'WMT_MS_Capabilities' in doc.tag:
        LOGGER.debug("replace proxy urls in wms 1.1.1")
        for element in doc.findall('.//OnlineResource[@xlink:href]', namespaces=ns):
            parsed_url = urlparse.urlparse(element.get('{http://www.w3.org/1999/xlink}href'))
            new_url = url
            if parsed_url.query:
                new_url += '?' + parsed_url.query
            element.set('{http://www.w3.org/1999/xlink}href', new_url)
        xml = etree.tostring(doc)
    # wms 1.3.0 onlineResource
    elif 'WMS_Capabilities' in doc.tag:
        LOGGER.debug("replace proxy urls in wms 1.3.0")
        for element in doc.findall('.//{http://www.opengis.net/wms}OnlineResource[@xlink:href]', namespaces=ns):
            parsed_url = urlparse.urlparse(element.get('{http://www.w3.org/1999/xlink}href'))
            new_url = url
            if parsed_url.query:
                new_url += '?' + parsed_url.query
            element.set('{http://www.w3.org/1999/xlink}href', new_url)
        xml = etree.tostring(doc)
    # wps operations
    elif 'Capabilities' in doc.tag:
        for element in doc.findall('ows:OperationsMetadata//*[@xlink:href]', namespaces=ns):
            element.set('{http://www.w3.org/1999/xlink}href', url)
        xml = etree.tostring(doc)
    elif prev_url:
        xml = xml.decode('utf-8', 'ignore')
        xml = xml.replace(prev_url, url)
    return xml



import colander
from cornice import Service
from cornice.validators import colander_body_validator

from twitcher.__version__ import __version__


import logging
LOGGER = logging.getLogger("TWITCHER")

# Rest API Services
services = Service(name='services',
                   path='/services',
                   permission='view',
                   description="List, add or clear services")

service = Service(name='service',
                  path='/services/{name}',
                  permission='view',
                  description="Get or remove a service item")


# Register service request schema
class ServicesPostBodySchema(colander.MappingSchema):
    name = colander.SchemaNode(colander.String(),
                               description='Service name')
    url = colander.SchemaNode(colander.String(),
                              description='Service URL')
    type = colander.SchemaNode(colander.String(),
                               missing=colander.drop, default='WPS',
                               description='Service Type')
    purl = colander.SchemaNode(colander.String(),
                               missing=colander.drop, default='',
                               description='Service public URL')
    verify = colander.SchemaNode(colander.Boolean(),
                                 missing=colander.drop, default=True,
                                 description='Verify HTTPS connection')
    auth = colander.SchemaNode(colander.String(),
                               missing=colander.drop, default='token',
                               description='Authentication method')


# Create our cornice service views
class TwitcherAPI(object):
    """Twitcher API defined with OpenAPI."""

    @staticmethod
    @services.get(tags=['services', 'list'])
    def list_services(request):
        """Returns a list of registered services."""
        return request.owsregistry.list_services()

    @staticmethod
    @services.delete(tags=['services', 'clear'])
    def clear_services(request):
        """Clear all services."""
        return request.owsregistry.clear_services()

    @staticmethod
    @services.post(tags=['services', 'register'],
                   validators=(colander_body_validator, ),
                   schema=ServicesPostBodySchema())
    def register_service(request):
        """Register a service."""
        LOGGER.debug("request validated={}".format(request.validated))
        return request.owsregistry.register_service(**request.validated)

    @staticmethod
    @service.get(tags=['service', 'get'])
    def get_service(request):
        """Get registered service."""
        return request.owsregistry.get_service_by_name(name=request.matchdict['name'])

    @staticmethod
    @service.delete(tags=['service', 'unregister'])
    def unregister_service(request):
        """Remove registered service."""
        return request.owsregistry.unregister_service(name=request.matchdict['name'])


def includeme(config):
    config.include('twitcher.basicauth')
    config.include('cornice')
    config.include('cornice_swagger')
    config.include('twitcher.oauth2')
    config.include('twitcher.owsregistry')
    # Create views to serve our OpenAPI spec
    config.cornice_enable_openapi_view(
        api_path='/__api__',
        # permission='view',
        title='Twitcher API',
        description="OpenAPI documentation",
        version=__version__
    )
    # Create views to serve OpenAPI spec UI explorer
    config.cornice_enable_openapi_explorer(
        api_explorer_path='/api-explorer',
        # permission='view'
    )
